﻿using System;
using System.Numerics;

namespace FastRng;

/// <summary>
/// Provides some mathematical function, which are not available within in .NET itself.
/// </summary>
public static class IntegerMathTools<TNum> where TNum : IBinaryInteger<TNum>, IComparisonOperators<HugeInteger<TNum>, TNum, HugeInteger<TNum>>, IMultiplyOperators<HugeInteger<TNum>, TNum, HugeInteger<TNum>>
{
    private static readonly TNum CONST20 = TNum.CreateChecked(20);
    
    /// <summary>
    /// The mathematical factorial function for integer numbers.
    /// </summary>
    /// <param name="x">The value, for which you want to know the factorial.</param>
    /// <exception cref="ArgumentOutOfRangeException">Throws, when x is greater than 20. Due to limitations of 64bit ulong type.</exception>
    public static ulong Factorial(TNum x)
    {
        if (x > CONST20)
            throw new ArgumentOutOfRangeException(nameof(x), $"Cannot compute {x}!, since ulong.max is 18_446_744_073_709_551_615.");
            
        var accumulator = new HugeInteger<TNum>(1);
        for (var factor = TNum.One; factor <= x; factor++)
            accumulator *= factor;

        return accumulator;
    }
}